//
//  Copyright (C) by S.Tomabechi 1999
//  All rights reserved.
//
//  Sparse matrix class over GF(2)
//

#include "SMatrix.hpp"
#ifdef __BORLANDC__
# include <mem.h>
#endif
#include <algorithm>
#include <vector>
#include "TBit.hpp"

using namespace std ;

const unsigned long SparseMatrix::isHeavy=0xFFFFFFFF ;
const unsigned long SparseMatrix::isLight=0xFFFFFFF8 ;
const unsigned long SparseMatrix::isDelete=0xFFFFFFF0 ;

//
// constructor (alloc pointers)
//
SparseMatrix::SparseMatrix(const unsigned long _nrow,const unsigned long _ncol)
{
    nrow=_nrow ;
    ncol=_ncol ;
    mat = new unsigned long*[nrow] ;
    memset(mat,0x00,nrow*sizeof(unsigned long*)) ;
    w_row=new unsigned long[nrow] ;
    memset(w_row,0x00,nrow*sizeof(unsigned long)) ;
    w_col=new unsigned long[ncol] ;
    memset(w_col,0x00,ncol*sizeof(unsigned long)) ;
}

//
//  destructor
//
SparseMatrix::~SparseMatrix()
{
    if( mat ) {
        for( unsigned long j=0 ; j<nrow ; j++ ) {
            if( mat[j]  ) delete [] mat[j] ;
        }
        delete [] mat ;
    }
    if( w_row ) delete [] w_row ;
    if( w_col ) delete [] w_col ;
}

//
//  indices must be sorted in ascending order
//  CfbNX܂܂Ă邩̓`FbNȂ
//
int SparseMatrix::SetRowVector(const unsigned long irow,const unsigned long nelem,
                                const unsigned long* index)
{
    if( mat[irow] ) delete [] mat[irow] ;
    if( nelem!=0 ) {
        mat[irow]=new unsigned long[nelem] ;
        memcpy(mat[irow],index,nelem*sizeof(unsigned long)) ;
    }
    else mat[irow]=new unsigned long[1] ;
    w_row[irow]=nelem ;
    for( unsigned long j=0 ; j<nelem ; j++ ) {
        w_col[index[j]]++ ;
    }
    return 1 ;
}


//
// transposed matrix
//
#if 0
// copy constructor
SparseMatrix::SparseMatrix(const SparseMatrix& m)
{
    unsigned long j ;
    nrow=m.nrow ;
    ncol=m.ncol ;
    mat=new unsigned long*[nrow] ;
    w_row=new unsigned long[nrow] ;

    w_col=new unsigned long[ncol] ;

    for( j=0 ; j<nrow ; j++ ) {
        mat[j]=new unsigned long[m.w_row[j]] ;
        w_row[j]=m.w_row[j] ;
        memcpy(mat[j],m.mat[j],w_row[j]*sizeof(unsigned long)) ;
    }
    for( j=0 ; j<ncol ; j++ ) {
        w_col[j]=m.w_col[j] ;
    }

}

// transposed matrix
SparseMatrix t(const SparseMatrix& m)
{
    unsigned long j,k ;
    SparseMatrix tm(m.ncol,m.nrow) ;
    for( j=0 ; j<tm.nrow ; j++ ) tm.mat[j]=new unsigned long[m.w_col[j]] ;

    unsigned long new_row,new_col ;

    for( j=0 ; j<m.nrow ; j++ ) {
        for( k=0 ; k<m.w_row[j] ; k++ ) {
            new_row=m.mat[j][k] ;    // m  column ԍ
            tm.mat[new_row][tm.w_row[new_row]]=j ;
            tm.w_row[new_row]++ ;
        }
        tm.w_col[j]=m.w_row[j] ;
    }

    return tm ;
}

//
// product of sparse matrices
//
SparseMatrix    SparseMatrix::operator*(const SparseMatrix& m)
{
    SparseMatrix p(nrow,m.ncol) ;
    unsigned long j,k,n ;
    unsigned long c ;
    vector< unsigned long > v ;
    vector< unsigned long >::iterator it,start,end ;

    for( j=0 ; j<nrow ; j++ ) {
        v.resize(0) ;
        for( k=0 ; k<m.ncol ; k++ ) {
            c=0 ;                   // 1 ̏o񐔂JEg
            if( w_row[j]!=0 ) {
                for( n=0 ; n<w_row[j] ; n++ ) {
                    start=m.mat[mat[j][n]] ;         //  mat[j][n] column
                    end=m.mat[mat[j][n]]+m.w_row[mat[j][n]]  ;
                    it=find(start,end,k) ;
                    if( it<end ) c++ ;
                }
                if( c&1 ) {
                    v.push_back(k) ;
                    p.w_col[k]++ ;
                }
            }
        }
        p.mat[j]=new unsigned long[v.size()] ;
        p.w_row[j]=v.size() ;
        for( n=0 ; n<v.size() ; n++ ) p.mat[j][n]=v[n] ;
    }

    return p ;
}




//
// assignment operetator
//
SparseMatrix& SparseMatrix::operator=(const SparseMatrix& m)
{
    if( this==&m ) return *this ;

    unsigned long j ;
    if( mat ) {
        for( j=0 ; j<nrow ; j++ ) {
            if( mat[j] ) delete [] mat[j] ;
        }
        delete [] mat ;
    }
    if( w_row ) delete [] w_row ;
    if( w_col ) delete [] w_col ;

    nrow=m.nrow ;
    ncol=m.ncol ;
    mat = new unsigned long*[nrow] ;
    w_row=new unsigned long[nrow] ;
    w_col=new unsigned long[ncol] ;
    for( j=0 ; j<nrow ; j++ ) {
        mat[j]=new unsigned long[m.w_row[j]] ;
        w_row[j]=m.w_row[j] ;
        memcpy(mat[j],m.mat[j],w_row[j]*sizeof(unsigned long)) ;
    }
    for( j=0 ; j<ncol ; j++ ) w_col[j]=m.w_col[j] ;
    return *this ;
}

::ostream& operator<<(::ostream& os,const SparseMatrix& m)
{
    unsigned long j,k,l ;
    for( j=0 ; j<m.nrow ; j++ ) {
        os << endl ;
        if( m.w_row[j]==0 ) {
            for( k=0 ; k<m.ncol; k++ ) os << "0" ;
        }
        else {
            for( k=l=0 ;;) {
                for( ; k<m.mat[j][l] ; k++ )    os << "0" ;
                os << "1" ;
                l++ ;
                k++ ;
                if( l>=m.w_row[j] ) {
                    for( ; k<m.ncol ; k++ ) os << "0" ;
                    break ;
                }
            }
        }

    }
    return os ;
}

#endif

// multiplication vm=v*m
void Mul(const unsigned long* v,const SparseMatrix& m,unsigned long* vm)
{
    unsigned long j,k ;
    unsigned long* w ;

    for( j=0 ; j<(m.ncol+31)/32 ; j++ ) vm[j]=0 ;

    w=new unsigned long[(m.ncol+31)/32] ;
    for( j=0 ; j<m.nrow ; j++ ) {
        if( BitIndex(v,j) ) {
            memset(w,0x00,((m.ncol+31)/32)*sizeof(unsigned long)) ;
            for( k=0 ; k<m.w_row[j] ; k++ ) BitOn(w,m.mat[j][k]) ;
            for( k=0 ; k<(m.ncol+31)/32 ; k++ ) vm[k]^=w[k] ;
        }
    }

    delete [] w ;
}

// multiplication vm=m*v
void Mul(const SparseMatrix& m,const unsigned long* v,unsigned long* mv)
{
    unsigned long j,k,n ;

    for( j=0 ; j<(m.nrow+31)/32 ; j++ ) mv[j]=0 ;

    for( j=0 ; j<m.nrow ; j++ ) {
        n=0 ;
        for( k=0 ; k<m.w_row[j] ; k++ ) {
            if( BitIndex(v,m.mat[j][k]) ) n++ ;
        }
        if( n&1 ) BitOn(mv,j) ;
    }

}


//
// debug print
//
void SparseMatrix::statics(::ostream& os) const
{
    unsigned long j ;
    unsigned long nentry ;
    for( nentry=0,j=0 ; j<nrow ; j++ ) nentry+=w_row[j] ;
    os << endl << nrow << "*" << ncol    << " matrix " ;
    os << double(nentry)/nrow << " entries / row " ;
    os << endl << "nonzero entry " << nentry
         << " memory " << (nentry*4+1023)/1024 << "K " ;
}


unsigned long SparseMatrix::NumberOfNonZeroEntry(void) const
{
    unsigned long nentry,j ;
    for( nentry=0,j=0 ; j<nrow ; j++ ) nentry+=w_row[j] ;
    return nentry ;
}

//
//  ߂l͒PɃfobOpɎgĂ邾
//
unsigned long SparseMatrix::DeleteColumn(unsigned long* new_col)
{
    unsigned long j ;
    long k ;
    unsigned long delcol(0) ;

    for( j=0 ; j<nrow ; j++ ) {
        if( w_row[j]==0 ) continue ;
        for( k=0 ; k<w_row[j] ; k++ ) {
            mat[j][k]=new_col[mat[j][k]] ;      // new column number
        }
        QSortULA((unsigned long)w_row[j],4,0,mat[j]) ;
        for( k=w_row[j]-1 ; k>=0 ; k-- ) {      // discard entries to be deleted
            if( mat[j][k]<isDelete ) {
                w_row[j]=k+1 ;
                break ;
            }
#if 1
if( mat[j][k]==isDelete ) delcol++ ;
#endif

        }
        if( k<0 ) {
            w_row[j]=0 ;
            delete [] mat[j] ;
            mat[j]=NULL ;
        }
    }

    for( k=0,j=0 ; j<ncol ; j++ ) {
        if( new_col[j]<isDelete ) {
            w_col[k]=w_col[j] ;               // change column weight
            k++ ;
        }
    }

    ncol=k ;
#if 1
return delcol ;
#endif
}


void read(::istream& is,SparseMatrix& m,
    const unsigned long maxrow,const unsigned long maxcol)
{
    unsigned long j,k,l ;
    m.~SparseMatrix() ;
    m.nrow=maxrow ;
    m.ncol=maxcol ;
    m.mat=new unsigned long*[m.nrow] ;
    m.w_row=new unsigned long[m.nrow] ;
    m.w_col=new unsigned long[m.ncol] ;
    memset(m.w_col,0x00,m.ncol*sizeof(unsigned long)) ;
    for( j=0 ; j<m.nrow ; j++ ) {
        is.read((char*)&k,sizeof(unsigned long)) ;
        if( k==0 ) l=1 ;
        else l=k ;
        m.mat[j]=new unsigned long[l] ;
        is.read((char*)m.mat[j],k*sizeof(unsigned long)) ;
        m.w_row[j]=(unsigned long)k ;
        for( l=0 ; l<k ; l++ ) {
            if( maxcol<m.mat[j][l] ) {
                m.w_row[j]=l ;
                break ;
            }
            else m.w_col[m.mat[j][l]]++ ;
        }
    }

}

// asǂŁA̓]us쐬
void readTransposed(::istream& is,SparseMatrix& m,
    const unsigned long maxrow,const unsigned long maxcol)
{
    unsigned long j,k,l,n ;
    unsigned long ipos=is.tellg() ;                         // t@C̈ʒu
    m.~SparseMatrix() ;
    m.nrow=maxrow ;
    m.ncol=maxcol ;
    m.mat=new unsigned long*[m.nrow] ;
    m.w_row=new unsigned long[m.nrow] ;
    m.w_col=new unsigned long[m.ncol] ;

    memset(m.w_row,0x00,m.nrow*sizeof(unsigned long)) ;
    memset(m.w_col,0x00,m.ncol*sizeof(unsigned long)) ;
    for( j=0 ; j<m.ncol ; j++ ) {
        is.read((char*)&k,sizeof(unsigned long)) ;
        for( l=0 ; l<k ; l++ ) {
            is.read((char*)&n,sizeof(unsigned long)) ; // n  ԍ
            if( n<maxrow ) {
                m.w_row[n]++ ;
                m.w_col[j]++ ;
            }

        }

    }

    for( j=0 ; j<m.nrow ; j++ ) {
        m.mat[j]=new unsigned long[m.w_row[j]] ;
    }
    memset(m.w_row,0x00,m.nrow*sizeof(unsigned long)) ;
    is.seekg(ipos) ;
    for( j=0 ; j<m.ncol ; j++ ) {
        is.read((char*)&k,sizeof(unsigned long)) ;
        for( l=0 ; l<k ; l++ ) {
            is.read((char*)&n,sizeof(unsigned long)) ; // n  ԍ
            if( n<maxrow ) {
                m.mat[n][m.w_row[n]]=j  ;
                m.w_row[n]++ ;
            }
        }

    }

}

// write sparse matrix as binary record
void write(::ostream& os,const SparseMatrix& m)
{
    unsigned long j,k ;
    for( j=0 ; j<m.nrow ; j++ ) {
        k=m.w_row[j] ;
        os.write((char*)&k,sizeof(unsigned long)) ;
        os.write((char*)m.mat[j],k*sizeof(unsigned long)) ;
    }

}

// ߂ĺAsingleton 폜܂̍s
// old_row ́AƂ̍sԍ
unsigned long SparseMatrix::DeleteSingleton(unsigned long* old_row)
{
    unsigned long j,k,l ;
    bool bChange ;

    do  {
        for( bChange=false, j=0 ; j<nrow ; j++ ) {
            for( k=0 ; k<w_row[j] ; k++ ) {
                if( w_col[mat[j][k]]==1 ) {           // singleton ?
                    bChange=true ;
                    for( l=0 ; l<w_row[j] ; l++ ) {
                        w_col[mat[j][l]]-- ;          // decrease column weight
                    }
                    delete [] mat[j] ;                // delete row
                    mat[j]=NULL ;
                    w_row[j]=0 ;
                }
            }
        }
    } while( bChange ) ;

//    old_row=new unsigned long[nrow] ;
    for( k=0,j=0 ; j<nrow ; j++ ) {
        if( w_row[j]!=0 ) {
//            old_row[k]=j ;
            old_row[k]=old_row[j] ;
            w_row[k]=w_row[j] ;
            mat[k]=mat[j] ;
            k++ ;
        }
    }
    l=nrow ;
    for( j=k ; j<nrow ; j++ ) { mat[j]=NULL ; w_row[j]=0 ; }
    nrow=k ;
    return l ;  // #rows before deleting singletons
}

//
//  row weight  0 ̍s폜
//
void SparseMatrix::DeleteNullRow(unsigned long* old_row)
{
    unsigned long j,k ;
    for( k=0,j=0 ; j<nrow ; j++ ) {
        if( w_row[j]!=0 ) {
            w_row[k]=w_row[j] ;
            old_row[k]=old_row[j] ;
            mat[k]=mat[j] ;
            k++ ;
        }
    }
    for( j=k ; j<nrow ; j++ ) {
        mat[j]=NULL ;
        w_row[j]=0 ;
    }
    nrow=k ;
}

